# -*- coding: utf-8 -*-
"""
app.networks.views
~~~~~~~~~~~~~~~~~~

This module implements the networks blueprint.

:copyright: (c) 2017 European Spallation Source ERIC
:license: BSD 2-Clause, see LICENSE for more details.

"""
import sqlalchemy as sa
from flask import (Blueprint, render_template, jsonify, session,
                   redirect, url_for, request, flash, current_app)
from flask_login import login_required
from .forms import HostForm
from ..extensions import db
from ..decorators import login_groups_accepted
from .. import models

bp = Blueprint('networks', __name__)


@bp.route('/hosts')
@login_required
def list_hosts():
    return render_template('networks/hosts.html')


@bp.route('/hosts/create', methods=('GET', 'POST'))
@login_groups_accepted('admin', 'create')
def create_host():
    # Try to get the network_id from the session
    # to pre-fill the form with the same network
    try:
        network_id = session['network_id']
    except KeyError:
        # No need to pass request.form when no extra keywords are given
        form = HostForm()
    else:
        form = HostForm(request.form, network_id=network_id)
    if form.validate_on_submit():
        network_id = form.network_id.data
        host = models.Host(ip=form.ip.data,
                           network_id=form.network_id.data,
                           name=form.name.data,
                           description=form.description.data or None,
                           mac_id=form.mac_id.data or None)
        current_app.logger.debug(f'Trying to create: {host!r}')
        db.session.add(host)
        try:
            db.session.commit()
        except sa.exc.IntegrityError as e:
            db.session.rollback()
            current_app.logger.warning(f'{e}')
            flash(f'{e}', 'error')
        else:
            flash(f'Host {host} created!', 'success')
        # Save network_id to the session to retrieve it after the redirect
        session['network_id'] = host.network_id
        return redirect(url_for('networks.create_host'))
    return render_template('networks/create_host.html', form=form)


@bp.route('/_retrieve_hosts')
@login_required
def retrieve_hosts():
    data = [(host.name,
             host.ip,
             host.description,
             str(host.mac),
             host.network.vlan_name) for host in models.Host.query.all()]
    return jsonify(data=data)


@bp.route('/_retrieve_available_ips/<network_id>')
@login_required
def retrieve_available_ips(network_id):
    try:
        network = models.Network.query.get(network_id)
    except sa.exc.DataError:
        current_app.logger.warning(f'Invalid network_id: {network_id}')
        data = []
    else:
        data = [str(address) for address in network.available_ips()]
    return jsonify(data=data)
